package se.paxport.yggejonssoncrazysnake;

import se.citerus.crazysnake.*;
import se.paxport.yggejonssoncrazysnake.util.Board;
import se.paxport.yggejonssoncrazysnake.util.SnakeImpl;

import java.util.*;

import static org.junit.Assert.fail;

public class AbstractPaxportTest {

    private final SnakeImpl currentSnake = new SnakeImpl(new BrainId(Paxport.class.getName()));

    private final SnakeImpl opponentSnake = new SnakeImpl(new BrainId("Opponent"));

    private final Map<BrainId, Snake> snakeMap = new HashMap<BrainId, Snake>();

    private final HashSet<BrainId> brainIds = new HashSet<BrainId>();

    private Board board;

    private HeatMeta heatMeta;

    protected void initBoard(String[] boardString) {
        board = new Board(boardString, currentSnake, opponentSnake);
        heatMeta = new HeatMeta(50, 10, board.getDimension());

        snakeMap.put(currentSnake.getId(), currentSnake);
        snakeMap.put(opponentSnake.getId(), opponentSnake);

        brainIds.add(currentSnake.getId());
        brainIds.add(opponentSnake.getId());
    }

    protected void setupSnakes() {
        setupSnakes(0, 0);
    }

    protected void setupSnakes(int currentScore, int opponentScore) {
        currentSnake.setScore(currentScore);
        opponentSnake.setScore(opponentScore);
    }

    protected void setupSnakes(SnakeSetup current, SnakeSetup opponent, boolean updateSnakesOnBoard) {
        current.initSnake(currentSnake);
        opponent.initSnake(opponentSnake);
        if (updateSnakesOnBoard) {
            board.updateSnakePositions();
        }
    }

    protected Movement getNextMove() {
        Paxport snake = new Paxport();
        snake.init(brainIds, heatMeta);
        HeatState state = new HeatState(board.getSquares(), snakeMap, 0, 0);

        List<Position> currentPositions = getPositions('a');
        List<Position> opponentPositions = getPositions('A');
        if (currentPositions.size() != opponentPositions.size()) {
            throw new RuntimeException("Snakes not of same length");
        }

        int len = 1;
        Position prevCurrentPos = new Position(0, 0);
        Position prevOpponentPos = new Position(0, 0);
        for (int i = currentPositions.size() - 1; i > 0; --i) {
            currentSnake.setPosition(currentPositions.get(i));
            currentSnake.setLength(len);
            currentSnake.setDirection(calcDirection(currentPositions.get(i), prevCurrentPos));
            prevCurrentPos = currentPositions.get(i);
            opponentSnake.setPosition(opponentPositions.get(i));
            opponentSnake.setLength(len);
            opponentSnake.setDirection(calcDirection(opponentPositions.get(i), prevOpponentPos));
            prevOpponentPos = opponentPositions.get(i);
            snake.updateSnakesWithSegments(state);
            ++len;
        }
        currentSnake.setPosition(currentPositions.get(0));
        currentSnake.setLength(len);
        currentSnake.setDirection(calcDirection(currentPositions.get(0), prevCurrentPos));
        opponentSnake.setPosition(opponentPositions.get(0));
        opponentSnake.setLength(len);
        opponentSnake.setDirection(calcDirection(opponentPositions.get(0), prevOpponentPos));
        return snake.getNextMove(state);
    }

    private Direction calcDirection(Position nextPosition, Position prevPosition) {
        if (nextPosition.getX() > prevPosition.getX()) {
            return Direction.EAST;
        } else if (nextPosition.getX() < prevPosition.getX()) {
            return Direction.WEST;
        } else if (nextPosition.getY() > prevPosition.getY()) {
            return Direction.SOUTH;
        } else {
            return Direction.NORTH;
        }
    }

    private List<Position> getPositions(char c) {
        List<Position> positions = new ArrayList<Position>();
        while (true) {
            Position pos = getPosForChar(c++);
            if (pos == null) {
                break;
            }
            positions.add(pos);
        }
        return positions;
    }

    private Position getPosForChar(char c) {
        String[] boardString = board.getBoard();
        for (int y = 0; y < boardString.length; ++y) {
            for (int x = 0; x < boardString[y].length(); ++x) {
                if (boardString[y].charAt(x) == c) {
                    return new Position(x, y);
                }
            }
        }
        return null;
    }

    protected void assertMovement(Movement... allExpected) {
        Movement move = getNextMove();
        for (Movement expected : allExpected) {
            if (move == expected) {
                return;
            }
        }
        fail(String.format("Move %s not expected in %s", move, Arrays.asList(allExpected)));
    }

    protected class SnakeSetup {
        Direction direction;

        Position pos;

        int length;

        int score;

        protected SnakeSetup(Direction direction, Position pos, int length) {
            this(direction, pos, length, 0);
        }

        protected SnakeSetup(Direction direction, Position pos, int length, int score) {
            this.direction = direction;
            this.pos = pos;
            this.length = length;
            this.score = score;
        }

        private void initSnake(SnakeImpl snake) {
            snake.setDirection(direction);
            snake.setLength(length);
            snake.setPosition(pos);
            snake.setScore(score);
        }
    }
}
